#include <iostream>
#include <stdlib.h>

#include <x86intrin.h>

static const int N = 1000;
static const int STEPS = 1000000;

struct Players
{
    bool *isA;
    bool *isTolerant;
    int *payoff;
    bool **judgedGood;
};

static __m128i scatter(int v)
{
     __m128i rs = _mm_cvtsi32_si128(v);                                     // rs    = x x x x x x x x x x x x 0 0 0 v
    __m128i zero = _mm_xor_si128(rs, rs);                                   // zero  = 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0

    return _mm_shuffle_epi8(rs, zero);                                      // rs    = v v v v v v v v v v v v v v v v
}

int main(int argc, char **argv)
{
    (void) argc;
    (void) argv;

    // Créer N joueurs qui tous jugés comme étant bons
    int tolerant_players = 0;
    Players players;

    players.isA = new bool[N];
    players.isTolerant = new bool[N];
    players.payoff = new int[N];
    players.judgedGood = new bool*[N];

    for (int i=0; i<N; ++i) {
        players.isA[i] = rand() & 1;
        players.isTolerant[i] = rand() & 1;
        players.payoff[i] = 0;
        players.judgedGood[i] = new bool[N];

        if (players.isTolerant[i]) {
            tolerant_players += 1;
        }

        // Initialiser les jugements
        for (int j=0; j<N; ++j) {
            players.judgedGood[i][j] = rand() & 3;
        }
    }

    // Simuler le jeu
    for (int t=0; t<STEPS; ++t) {
        // Choisir deux joueurs
        int player_a = rand() % N;
        int player_b = rand() % N;

        // Savoir si le joueur A va coopérer avec le joueur B
        bool cooperate = false;

        if (!players.isTolerant[player_a]) {
            cooperate = (players.isA[player_a] == players.isA[player_b]);
        } else {
            cooperate = (players.judgedGood[player_b][player_a] | players.judgedGood[player_a][player_a]);
        }

        cooperate &= rand() & 7;       // Une chance sur 8 que la coopération échoue

        if (cooperate) {
            players.payoff[player_a] -= 2;
            players.payoff[player_b] += 3;
        }

        // A va devenir tolérant ou intolérant suivant comment il se compare avec B
        if (players.isTolerant[player_a] != players.isTolerant[player_b] &&
            players.payoff[player_a] < players.payoff[player_b])
        {
            players.isTolerant[player_a] = players.isTolerant[player_b];

            tolerant_players += (players.isTolerant[player_b] ? 1 : -1);
        }

        // A va maintenant être jugé par tous les autres joueurs
        __m128i first_is_a_v = scatter(players.isA[player_a]);
        __m128i second_is_a_v = scatter(players.isA[player_b]);
        __m128i cooperate_v = scatter(cooperate);

        for (int other=0; other<N-16; other += 16) {
            __m128i other_is_tolerant_v = _mm_load_si128((__m128i *)&players.isTolerant[other]);
            __m128i other_is_a_v = _mm_load_si128((__m128i *)&players.isA[other]);
            __m128i second_considered_good_v = _mm_load_si128((__m128i *)&players.judgedGood[player_b][other]);

            // Les joueurs tolérant jugent les autres joueurs bons si "cooperate && players.judgedGood[player_b][other]"
            __m128i tolerant_judgment = _mm_and_si128(cooperate_v, second_considered_good_v);

            // Les joueurs intolérants jugent les autres bons selon tout un tas de conditions
            __m128i other_a_different_from_first_a = _mm_xor_si128(other_is_a_v, first_is_a_v);
            __m128i second_a_different_from_first_a = _mm_xor_si128(second_is_a_v, first_is_a_v);

            __m128i force_bad = _mm_or_si128(                               // Joueur forcément mauvais si :
                other_a_different_from_first_a,                             //      il est dans une classe différente de "other" (le joueur qui juge) OU
                _mm_and_si128(cooperate_v, second_a_different_from_first_a) //      il a coopéré avec un joueur d'une autre classe
            );
            __m128i force_good = _mm_andnot_si128(                          // Joueur forcément bon si :
                cooperate_v,                                                //      il n'a pas coopéré (andnot inverse son premier paramètre) ET
                second_a_different_from_first_a                             //      l'autre joueur était d'une classe différente
            );

            // Fusionner le jugement tolérant et intolérant
            __m128i judgment = tolerant_judgment;

            judgment = _mm_or_si128(        // Bon si on considère le joueur bon, ou si on est intolérant et qu'on le force à bon
                _mm_andnot_si128(
                    other_is_tolerant_v,
                    force_good
                ),
                judgment
            );
            judgment = _mm_andnot_si128(    // Mauvais si on considère le joueur mauvais ou si on est intolérant et qu'on le force à mauvais
                _mm_andnot_si128(
                    other_is_tolerant_v,
                    force_bad
                ),
                judgment
            );

            _mm_store_si128((__m128i *)&players.judgedGood[player_a][other], judgment);
        }
    }

    std::cout << "Nombre de joueurs tolérants à la fin du jeu : " << tolerant_players << std::endl;

    // Supprimer les données
    for (int i=0; i<N; ++i) {
        delete[] players.judgedGood[i];
    }

    delete[] players.isA;
    delete[] players.isTolerant;
    delete[] players.payoff;
    delete[] players.judgedGood;
}